/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <ctype.h>
#include <dirent.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>  

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "dict.hpp"


text sol_pair::toText() {
  text t = key;
  t.cat(" : ");
  t.cat(value->toText().c_str());
  return t;
}

char *sol_pair::toStr() {
  return value->toStr();
}


/** dict.cpp
  File of the dict and dict.
  */


extern void errorMsg(char *);
int dict::nomatch()   { errorMsg("dict type mismatch"); return 0; }

dyn dict::nullelement = dyn(false);

/* Constructors */

dict::dict() {
  indexer = 0;
}


 /**
   Build an associative dict from an dict of strings of chars.
   Indexes are generated from 0 to n.
   */


dict::dict(dyn x) {
    indexer = 0;
    push(x);
}

dict::dict(const dict &d) { set(d); }

dict::dict(dict *d) { set(d); }


dict::dict(std::map<text, dyn> &a) { set(a); }
dict::dict(std::map<text, dyn> *a) { set(&a); }


dict::~dict()
{
  keyList.clear();
  data.clear();
}


void dict::clear()
{
  keyList.clear();
  data.clear();
  indexer = 0;
}

void dict::pushKey(const text &s) {
    keyList.push_back(s);
}

void dict::set(const dict &a)  {
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  indexer = 0;
}
void dict::set(dict *a) {
  keyList.assign(a->keyList.begin(), a->keyList.end());
  data.assign(a->data.begin(), a->data.end());
  indexer = 0;
}

void dict::set(std::map<text, dyn> &a) {
  clear();
  for(auto const & x : a) {
    push(x.first, x.second);
  }    
}


void dict::set(dyn d) {
  if(d.flag != TYPE_DICT) { 
      clear(); 
      push(d); 
      return; 
  }
  set(d.d);
}


dyn dict::get(int n)
{
  char buffer[64];
  sprintf(buffer, "%d", n);
  return get(buffer);
}

dyn dict::get(const text &a)
 {
  for (unsigned int i = 0; i < keyList.size(); i++)
   {
     if(strcmp(a.data, keyList[i].data) == 0) return data[i];
   }
  return dyn(false);
 }

dyn dict::get(const char *a)
{
  for (unsigned int i = 0; i < keyList.size(); i++)
   {
     if(strcmp(a, keyList[i].data) == 0) return data[i];
   }
  return dyn(false);
}


dyn &dict::at(int n)  {
   int k = findKey(n);
   if(k == -1)  {
     dyn::error = true;
     return nullelement;
   }
   return data[k];
 }
 
 dyn &dict::at(const char *s) { return at(text(s)); }
 
 dyn &dict::at(text t) {
   int k = findKey(t);
   if(k == -1)  {
     dyn::error = true;
     return nullelement;
   }
   return data[k];
 }
 
dyn &dict::at(dyn d) { 
    if(d.isText())  return at(d.toText()); 
    return at(d.toInt());
} 

void dict::setAt(int idx, dyn *a)
 {
   int k = findKey(idx);
   if(k != -1)
   {
     // the dyn held at k, in current dict, replaced by a
     data[k].set(*a);
     return;
   }
   pushAdd(idx, *a);
 }

void dict::setAt(int idx, dyn a)
 {
   int k = findKey(idx);
   if(k != -1)
   {
     data[k].set(a);
     return;
   }
   pushAdd(idx, a);
 }


void dict::setAt(const text &k, dyn a) { setAt(k, (dyn *) &a); }
void dict::setAt(const text &k, dyn *v)
 {
   int i = findKey(k.data);
   if(i == -1)
   {
     push(k,v);
     return;
   }
   data[i].set(v);
 }


void dict::add(const dict &a)
{
  const dict *x = &a;
  add((dict *) x);
}

void dict::add(dict *x) {
  int k;
  for(int i = 0; i < x->size(); i++)  {
    k = findKey(x->keyList[i]);
    if(k != -1) continue;
      keyList.push_back(x->keyList[i]);
      data.push_back(x->data[i]);
  }
}


void dict::clone(dict *y)
 {
   y->keyList.assign(keyList.begin(), keyList.end());
   y->data.assign(data.begin(), data.end());
   y->indexer = 0;
 }

//===== OPERATORS

void dict::operator=(const dict &a)
{
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  indexer = 0;
  return;
}

void dict::operator=(dict *a)  { set(a); }
void dict::operator=(dyn a)    { set(a); }
// work in progress
void dict::operator=(std::map<text, dyn> &m) {
  for(auto k:m) {
      keyList.push_back(k.first);
      data.push_back(k.second);
  }
  indexer = 0;
  return;
}



 dict dict::operator+(dict *x)
 {
   dict y = dict(this);
   y.add(x);
   return y;
 }

 dict dict::operator+(dyn x)
 {
   dict y = dict(this);
   y.push(x);
   return y;
 }

// clone current array minus item in array b

dict dict::operator-(dict *b) {
   dyn v;
   int i,j;
   int l = b->size();
   dict y = dict();
   text k;

   for(i = 0; i < size(); i++) {
     k = keyList[i];
     v.set(data[i]);
     for(j = 0; j  < l; j++)  {
       if(v == &b->data[j]) goto skip;
     }
     y.push(k, v);
     skip:;
   }
   return y;
};

dict dict::operator-(dyn x) {
   dict y = dict();
   for(int i = 0; i < size(); i++)  {
     if(data[i] != x)  {
       y.push(keyList[i], data[i]);
     }
   }
   return y;
}

// dict dict::operator&(dict a) { return operator&(a); }
 dict dict::operator&(dict *b)
 {
   dyn v;
   int i,j;
   int l = b->size();
   dict *y = new dict();
   text k;

   for(i = 0; i < size(); i++)
   {
     k = keyList[i];
     v.set(data[i]);
     for(j = 0; j  < l; j++)
     {
       if(v == b->data[j])
       {
         if(text::isNumber(k.data))
           y->push(v);
         else
         {
           y->keyList.push_back(k);
           y->data.push_back(v);
         }
       }
     }
   }
   y->pack();
   return *y;
 };


//dict dict::operator|(dict a){ return operator|(a); };

dict dict::operator|(dict *b)
{
  unsigned int i, l = size();
  dict *y = new dict();

  for(i = 0; i < l; i++)
  {
    y->pushUnic(keyList[i], data[i]);
  }
  y->add(b);
  y->pack();

  return y->unique();
}

//dict dict::operator^(dict a){ return operator|(a); };
dict dict::operator^(dict *b)
{
  int i;
  dict *y = new dict();

  for(i = 0; i < size(); i++)
  {
    if(b->findValue(data[i]) == -1)
      y->pushUnic(keyList[i], data[i]);
  }
  for(i = 0; i < b->size(); i++)
  {
    if(findValue(b->data[i]) == -1)
      y->pushUnic(b->keyList[i], b->data[i]);
  }
  return *y;
}

/*void dict::operator+=(const dict &x) { this->add(x); return *this; }*/
dict &dict::operator+=(dict *x)
{
   this->add(x);
   return *this;
}
dict &dict::operator+=(dyn x)
{
   //puts("by dict");
   push(x);
   return *this;
}


// void dict::operator-=(dict a) { operator-=(a); }
 void dict::operator-=(dict *x)
 {
   dict y = dict(this);
   set(y.operator-(x));
 }

 void dict::operator-=(dyn x) {
   //if(x.flag ==  TYPE_ARRAY) operator-=(x.a);
   dict y = dict(this);
   set(y - x);
 }

// void dict::operator&=(dict b) { operator&=(b); }
 void dict::operator&=(dict *b)
 {
   dict y = operator&(b);
   set(y);
 }
// void dict::operator|=(dict b) { operator|=(b); }
 void dict::operator|=(dict *b)
 {
   dict y = operator|(b);
   set(y);
 }
// void dict::operator^=(dict b) { operator^=(b); }
 void dict::operator^=(dict *b)
 {
   dict y = operator^(b);
   set(y);
 }




bool dict::operator==(dict *b)
 {
   int l = size();
   if(b->size() != l) return false;
   for(int i = 0; i < l; i++)
   {
     if(strcmp(keyList[i].data, b->keyList[i].data) != 0) return false;
     if(data[i] != b->data[i]) return false;
   }
   return true;
 }

// bool  dict::operator==(dict b) { return operator==(b); };
 bool  dict::operator==(bool x)  { return (size() != 0) == x; }

// bool  dict::operator!=(dict a)  { return !operator==(a); };
 bool  dict::operator!=(dict *a) { return !operator==(a); };
 bool  dict::operator!=(bool x) { return !operator==(x); }

// bool dict::operator<(dict a) { return operator<(a); };
 bool dict::operator<(dict *a)
 {
   int l = a->size();
   if(size() < l) return true;
   if(size() > l) return false;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return false;
     if(data[i] < a->data[i]) return true;
   }
   return false;
 }

// bool dict::operator>(dict a) { return operator>(a); };
 bool dict::operator>(dict *a)
 {
   int l = a->size();
   if(size() < l) return false;
   if(size() > l) return true;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return true;
     if(data[i] < a->data[i]) return false;
   }
   return false;
 }

bool dict::operator!() { return size() == 0; }

dict &dict::operator,(dyn x) { push(x); return *this; }

//===== util

 bool dict::in(const dyn &d)  {
  for(unsigned int i = 0; i < data.size(); i++)  {
   if(data[i] == d) return true;
  }
  return false;
 }

 bool dict::in(dyn *d)  {
    for(unsigned int i = 0; i < data.size(); i++)  {
        if(data[i] == d) return true;
    }
    return false;
 }

/* is the textual key already in dict? */
signed int dict::hasStrKey(const text &s) {
    if(!text::isNumber(s.data))  {
    for(unsigned i = 0; i < keyList.size(); i++)
        if(strcmp(s.data, keyList[i].data) == 0) return i;
    }
    return -1;
}

signed int dict::findKey(const text &s) {
    for(unsigned i = 0; i < keyList.size(); i++)   {
    if(strcmp(s.data, keyList[i].data) == 0) return i;
    }
    return -1;
}

signed int dict::findKey(const char *s)
{
  int l = size();
  for(int i = 0; i < l; i++)
  {
   if(strcmp(s, keyList[i].data) == 0) return i;
  }
  return -1;
}


signed int dict::findKey(int n) { return findKey(text(n)); }

signed int dict::findValue(dyn *d) { return findValue(*d); }
signed int dict::findValue(const dyn &d)
{
  for(int i = 0; i < size(); i++)
  {
   if(data[i] == d) return i;
  }
  return -1;
}

/*
  Find last value
*/

signed int dict::findLastValue(dyn *d) { return findLastValue(*d); }
signed int dict::findLastValue(const dyn &d)
{
  for(int i = size() - 1; i >= 0; i--)
  {
   if(data[i] == d) return i;
  }
  return -1;
}

 //===== iterator

 void dict::reset() { indexer = 0; }
 
 bool dict::outOfBound() { 
     return indexer >= size(); 
 }

 bool dict::bound()  { 
     return(++indexer < size()); 
 }
 
 bool dict::inBound() {
  if(indexer < size()) { 
      return true;
  }
  return false;
 }
 
 dyn dict::inc()  {
  if(++indexer < size()) { 
      return data[indexer]; 
  }
  indexer = size();
  return dyn(false);
 }

 dyn dict::dec() {
  if(--indexer >= 0) return data[indexer];
  indexer = 0;
  return dyn(false);
 }

 dyn dict::current() { return value(indexer); }

 dyn dict::value() { return value(indexer); }

 dyn dict::value(int i) {
   //std::cout << i << "/" << size() << " " << data[i].toStr() << std::endl;
   if(i >=0 && i < size())  return data[i];
   return dyn(false);
  }

 text dict::key() {
   if(indexer >= 0 && indexer < (int) keyList.size())  return keyList[indexer];
   return text();
  }
 text dict::key(int i)    {
   if(i >= 0 && i < (int) keyList.size())  return keyList[i];
   return text();
  }
 int dict::index()
 {
   if(indexer >=0 && indexer < (int) keyList.size()) return atoi(keyList[indexer].c_str());
   return -1;
 }

dyn dict::begin()
 {
   indexer = 0;
   if(data.size() > 0) return data.front();
   return dyn(false);
}
dyn dict::end()   {
   indexer = 1;
   if(data.size() > 0)
   {
     indexer = (int) data.size() -1;
     return data.back();
   }
   return dyn(false);
}

//===== stack

void dict::push(const dyn &a) {
   if(a.flag == TYPE_PAIR) push(a.p);
 }

 void dict::push(dyn *a) {
   if(a->flag == TYPE_PAIR) push(a->p->key, a->p->value);
 }

 void dict::push(sol_pair *p)
 {
    push(p->key, p->value);
    return;
 }

 void dict::push(const text &k, dyn v)
 {
    int i;
    if((i = findKey(k)) != -1)
     {
      data[i].set(v);
      return;
     }
    keyList.push_back(k);
    data.push_back(v);
 }

 void dict::pushAdd(const text &k, dyn v)  {
    keyList.push_back(k);
    data.push_back(v);
 }

 void dict::pushAdd(int k, dyn v)  {
    keyList.push_back(text(k));
    data.push_back(v);
 }


 void dict::pushUnic(const text &k, dyn v)  {
     if(findValue(v) == -1) {
        keyList.push_back(k);
        data.push_back(v);
        return;
     }
 }

 dyn dict::pop()  {
   dyn x = dyn(false);
   indexer = 0;
   if(!empty())
   {
     x.set(data.back());
     keyList.pop_back();
     data.pop_back();
   }
   return x;
 }

dyn dict::shift()
{
	if(empty())  return dyn(false);
	else
	{
		dyn *x = &data[0];
		erase(0);
		return *x;
	}
}

void dict::unshift(text t, dyn a)
{
    erase(t.data);
    keyList.insert(keyList.begin(), t);
    data.insert(data.begin(), a);
    pack();
}


void dict::insert(int n, dyn d) {
  if(n < 0) n = size() + n;
  if(n < 0) return;
  
  bool TA = (d.flag == TYPE_DICT);
  
  if(n >= size())   {
    if(TA) add(d.d);
    else  push(d);
    return;
  }

  if(TA)  {
   keyList.insert(keyList.begin() + n, d.d->keyList.begin(), d.d->keyList.end());
   data.insert(data.begin() + n, d.d->data.begin(), d.d->data.end());
  }
  else  {
   keyList.insert(keyList.begin() + n, text(65000));
   data.insert(data.begin() + n, d);
  }
  pack();
}

//===== RANGE

 void dict::pack()
 {
    return;
 }

 // erase from first to last included
 void dict::erase(int first, int last)
 {
    keyList.erase(keyList.begin() + first, keyList.begin() + last + 1);
    data.erase(data.begin() + first, data.begin() + last + 1);
 }

 void dict::erase(int pos)
 {
    keyList.erase(keyList.begin() + pos);
    data.erase(data.begin() + pos);
 }

 void dict::erase(const char *s)
{
   for(unsigned i = 0; i < keyList.size(); i++)
   {
     //printf("%s = %s:%s\n", s, keyList[i].data, data[i].toStr());
     if(strcmp(s, keyList[i].data) == 0)
     {
       keyList.erase(keyList.begin() + i);
       data.erase(data.begin() + i);
       return;
     }
   }
}


 dict dict::slice(int f, int l, int incl)
 {
   dict x;

   if(empty()) goto bye;
   l += incl;
   l--;
   if(l < 0) l += size();
   if(l < 0) goto bye;
   if(l >= size()) l = size() - 1;

   if(f < 0) f += size();
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= size()) goto bye;

   x.keyList.assign(keyList.begin() + f, keyList.begin() + l + 1);
   x.data.assign(data.begin() + f, data.begin() + l + 1);
   x.pack();
bye:
   return x;
 }

 void dict::splice(int b, int e)
 {
   if(b < 0) b = size() + b;
   if(b < 0) b = 0;
   if(e < 0) e = size() + e;
   if(e < 0) return;
   if(e >= size()) e = size() - 1;
   if(e < b) return;
   erase(b, e);
 }

 void dict::splice(int b, int e, int inc)
 {
   splice(b, e + inc - 1);
 }


 void dict::splice(int f)  { erase(f); }

 void dict::splice(int i, int last, dict a, int incl) { splice(i, last, &a, incl); }
 void dict::splice(int i, int last, dict *a, int incl)
 {
  if(empty()) { operator=(a); return; }
  if(last < 0) last = size() - last;
  if(i < 0) i += size();
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= size()) last = size() - 1;
  //cout << i << " " << last << " " << size() << std::endl;
  if(i >= size()) { add(a); return; }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i)
  {
    if(!a->empty()) insert(i, dyn(a));
    return;
  }
  if(a->empty()) { erase(i, last); return; }

  int j = 0;
  while((i <= last) && (j < a->size()))
  {
    keyList[i].set(a->keyList[j]);
    data[i].set(a->data[j]);
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a->size())
  {
   keyList.insert(keyList.begin() + i, a->keyList.begin() + j, a->keyList.end());
   data.insert(data.begin() + i, a->data.begin() + j, a->data.end());
  }
  pack();
 }


 void dict::splice(int i, int last, dyn a, int incl)
 {
  if(i < 0) i = size() - i;
  if(i < 0) i = 0;
  if(last < 0) last = size() - last;
  last += incl;
  last --;
  if(last >= size()) last = size() - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= size()) { push(a); return; }
  if(last == i)   { insert(i, a); }

  keyList[i].set(text("0"));
  data[i].set(a);
  erase(i + 1, last);
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

dict dict::unique()
{
  int i, l = size();
  dict y = dict();

  for(i = 0; i < l; i++)
  {
    if(y.findValue(data[i]) == -1)
      y.push(keyList[i], data[i]);
  }
  return y;
}


void dict::swapData(int a, int b)
{
  dyn x;
  x.set(data[a]);
  data[a].set(data[b]);
  data[b].set(x);
}

void dict::swapElem(int a, int b)
{
  text s = keyList[a];
  dyn x;
  x.set(data[a]);
  keyList[a].set(keyList[b]);
  data[a].set(data[b]);
  keyList[b].set(s);
  data[b].set(x);
}

void dict::keysSort(int first,int last)
{
 int f=first;
 int l=last;
 dyn item;
 item.set(keyList[(f + l) / 2]);

 do
 {
  while(keyList[f] < item) f++;
  while(item < keyList[l]) l--;
  if(f <= l)
  {
   swapElem(f, l);
   f++;
   l--;
  }
 } while(f <= l);

 if(first < l)  keysSort(first, l);
 if(f < last)   keysSort(f, last);
 return;
}

void dict::dynSort(int first,int last)
{
 int f=first;
 int l=last;
 dyn item;
 item.set(data[(f + l) / 2]);

 do
 {
  while(data[f] < item) f++;
  while(item < data[l]) l--;
  if(f <= l)
  {
   swapElem(f, l);
   f++;
   l--;
  }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void dict::dynRSort(int first, int last)
{
 dynSort(first, last);
 reverse();
 return;
}


void dict::sort(int x)  { x=0; dynSort(0, size() - 1); pack(); }
void dict::rSort(int x) { x=0; dynRSort(0, size()-1); }

void dict::kSort(int x) { 
    std::vector<text> k = keyList;
    std::vector<dyn> v;
    std::sort(k.begin(), k.end());
    for(text key : k) {
        for(int i = 0; i < keyList.size(); i++)  {
            if(strcmp(key.data, keyList[i].data) == 0) {
                v.push_back(data[i]);
            }
        }
    }
    keyList = k;
    data = v;
}


solarray dict::keys()
{
  solarray x = solarray();
  for(unsigned int i = 0; i < keyList.size(); i++) x.push(dyn(keyList[i]));
  return x;
}

solarray dict::values()
{
  solarray x = solarray();
  for(unsigned int i = 0; i < data.size(); i++) {
      dyn y = data[i];
      x.push(y);
  }
  return x;
}

dyn dict::min()
{
  double m = DBL_MAX;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x < m) m = x;
  }
  return dyn(m);
}

dyn dict::max()
{
  double m = 0;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x > m) m = x;
  }
  return dyn(m);
}

/*
  Functional map method
  */

dict dict::map(bool (*f)(dyn))
{
  int s = data.size();
  dict temp;
  for(int i = 0; i < s; i++)
  {
    if(f(data[i]))
      temp.push(data[i]);
  }
  return temp;
}

double dict::sum()
{
  double total = 0;
  dyn d;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) total += d.toReal();
    else
     if(d.flag == TYPE_TEXT)
     {
       total += d.t->toNumber();
     }
  }
  return total;
}

//===== OPERATIONS

text dict::join(const char *sep)
{
  int lsep, si, sum, offset;
  int i;

  lsep = strlen(sep);
  si = size();
  sum = 0;
  offset = 0;

  for(i = 0; i < si; i++)
  {
    if(data[i].flag == TYPE_TEXT)
    {
     sum+= data[i].t->length();
     sum+= lsep;
    }
  }

  char *nstr = (char *) new char[sum+1];
  for(i = 0; i < si; i++)
  {
   if(data[i].flag == TYPE_TEXT)
   {
     if(offset > 0)
     {
      strcpy(nstr + offset, sep);
      offset += lsep;
     }
     strcpy(nstr + offset, data[i].t->data);
     offset += data[i].t->length();
     nstr[offset] = '\0';
   }
  }
  nstr[offset] = '\0';
  return text(nstr);
}

dict dict::reverse()
{
  dict y = dict();
  for(int i = size() - 1; i >= 0; i--)
  {
    y.keyList.push_back(keyList[size() - i - 1]);
    y.data.push_back(data[i]);
  }
  return y;
}


std::ostream& operator<< (std::ostream& out, dict *d)
{
  return out << d->toText().data;
}


void dict::display() { display(0); }
void dict::display(int tab)
{
 int l = size();
 dyn x;
 std::cout << "dict(\n";
 for(int i = 0; i < l; i++)
 {
   //std::cout << "flag=" << data[i].flag << " " << data[i].toStr() << std::endl;
   for(int j = 0; j < tab; j++) std::cout << "     ";
   std::cout << "   " << keyList[i].data;
   std::cout << " : ";
   x.set(data[i]);
   if(x.flag == TYPE_ARRAY) x.a->display(tab + 1);
   else
    std::cout << x.toStr() << std::endl;
 }
 for(int j = 0; j < tab; j++) std::cout << "     ";
 std::cout << ")\n";
}

// Convert dict to string

char *dict::toString(char *s) {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t->data;
}

text dict::toText(char *s) {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t;
}


#ifdef DEBUG
void dict::image()
{
 int l = (int) size();
 text x = "image[" + text(l) + "] (\n";
 char buf[64];

 for(int i = 0; i < l; i++)
 {
   x += key(i);
   x += ":";
   x += value(i).toText();
   x += " (";
   x += value(i).toText().length();
   x += ") ";
   x += text::isNumber(keyList[i].data);
   x += " adr=";
   sprintf(buf, "%ul", &data[i]);
   x += text(buf);
   x += "\n";
 }
 x += ")";

 std::cout << x << std::endl;
}
#endif

std::ostream& operator<< (std::ostream& out, dict d)
{
  return out << d.toText().data;
}

void dict::load(const char *name) {
    std::ifstream input(name);
    this->clear();
	for(std::string line; getline(input, line); ) {
		push(line.c_str());
	}
	input.close();
	return;
}

bool dict::store(const char * name)
{
    std::ofstream f;
    f.open(name);
	for(int i = 0; i < size(); i++)	{
        f << data[i].toStr() << "\n";
	}
	f.close();
	return true;
}

dict *dict::to(const char *s) { return to(text(s)); }
dict *dict::to(text t) {
   int k = findKey(t);
   if(k == -1) {
     k = size();
     keyList.push_back(text(k));
     data.push_back(dyn(new dict()));
   }
   return data[k].toDict();
}

dict *dict::to(int n) {
  return to(text(n));
}

dyn & dict::operator[](const text &t) { return operator[](t.data); }
dyn & dict::operator[](const char *s)
{
  int k = size();
  for(int i = 0; i < k; i++)
  {
     if(strcmp(s, keyList[i].data) == 0)
     {
       return data[i];
     }
  }
  keyList.push_back(text(s));
  data.push_back(dyn(false));
  return *(data.begin() + k);
}


void dict::boundError(const text &t)
{
  std::cout << "Out of boundary " << t << std::endl;
}
