/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <dirent.h>

#include <string>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif
#include <math.h>

#include <iostream>
#include <vector>
#include <map>

#ifndef _RWSTD_NO_NAMESPACE
  using namespace std;
#endif


#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "dict.hpp"


/** \class dyn
   Base class for dynamic variables
 */


//================================= nil

nil::nil() { flag = TYPE_BOOL; b = false; }


bool nil::is()
{
  switch(flag)
  {
   case TYPE_BOOL:   return b == false;
   case TYPE_DOUBLE:
   case TYPE_NUMBER: return n == 0;
   case TYPE_INT:    return n == -1;   // for indices
   case TYPE_TEXT:   return (t == NULL) || (t->length() == 0);
   case TYPE_ARRAY:  return (a == NULL) || (a->size() == 0);
   case TYPE_DICT:  return (d == NULL) || (d->size() == 0);
   case TYPE_FILE:   return f == NULL;
   case TYPE_OBJECT: return o == NULL;
   default: break;
  }
  return true;
}

bool nil::is(bool x)    { return x == false; }
bool nil::is(int x)     { return x == -1; }
bool nil::is(natural x) { return x == 0; }
bool nil::is(double x)  { return x == 0; }
bool nil::is(text &x)   { return (x.length() == 0); }
bool nil::is(solarray &x)  { return (x.size() == 0); }
bool nil::is(dyn &x)    { return (x.isNil() == 0); }
bool nil::is(FILE *x)   { return x == NULL; }
bool nil::is(object *x) { return x == NULL; }

/*
   dyn object
   this is the static replacement for any dynamic variable
*/

dyn dyn::error = false;

bool dyn::isNil(dyn a) { return a.isNil(); }
bool dyn::isNil()
{
  switch(flag)
   {
     case TYPE_BOOL:   return !b;
     case TYPE_NUMBER: return n == -1;
     case TYPE_TEXT:   return t->length() == 0;
     case TYPE_OBJECT: return o == NULL;
     case TYPE_FILE:   return f == NULL;
     case TYPE_DICT:    return d->empty();
     case TYPE_ARRAY:  return a->empty();
     case TYPE_PAIR:   return p == NULL;
     case TYPE_NATURAL:return l == 0;
     default:          break;
   }
  return true;
 }

void errorMsg()        { cout << "Dyn error...\n"; }
void errorMsg(char *e) { cout << "Dyn error, " << e << "\n"; }

int dyn::nomatch(char c) {
  printf("Dyn error code %c, type mismatch, current #%u\n", c,  flag);
  return 0;
}

void unknown(char *op)      { printf("Dyn error on operator %s type mismatch.", op); }
int notYet()        { errorMsg("not yet implemented"); return 0; }
void errorMethod()  { errorMsg("this method can't apply to the data"); }
void unknow(int i)  { printf("Dyn error in set, bad type #%d\n", i); }

//================================ Assignment

//dyn::dyn(const text &x) { flag = TYPE_TEXT; t = new text(x); }

void dyn::set(bool x)          { flag = TYPE_BOOL; b = x; }
void dyn::set(unsigned char a) { flag = TYPE_NUMBER; n = (double) a; }
void dyn::set(char a)          { flag = TYPE_NUMBER; n = (double) a; }
void dyn::set(int a)           { flag = TYPE_NUMBER; n = (double) a; }
void dyn::set(unsigned int x)  { flag = TYPE_NUMBER; n = (double) x; }
void dyn::set(natural a)       { flag = TYPE_NUMBER; n = (double) a; }
void dyn::set(float a)         { flag = TYPE_NUMBER; n = (double) a; }
void dyn::set(double a)        { flag = TYPE_NUMBER; n = a; }
void dyn::set(FILE *x)         { flag = TYPE_FILE;   f = x; }
void dyn::set(void *v)         { flag = TYPE_OBJECT; o = v; }

void dyn::set(solarray *x)          { flag = TYPE_ARRAY; a = new solarray(x); }
void dyn::set(const solarray &x)    { flag = TYPE_ARRAY; a = new solarray(x);  }
void dyn::set(dict *x)              { flag = TYPE_DICT; d = new dict(x); }
void dyn::set(const dict &x)        { flag = TYPE_DICT; d = new dict(x);  }


void dyn::set(const char *s)     { flag = TYPE_TEXT;  t = new text(s); }
void dyn::set(const text &x)     { flag = TYPE_TEXT;  t = new text(x.data); }
void dyn::set(text *x)           { flag = TYPE_TEXT;  t = new text(x->data); }
void dyn::set(sol_pair *c)       { flag = TYPE_PAIR;  p = new sol_pair(c->key, c->value); }
void dyn::set(const dyn &x)  {
  switch(x.flag)  {
     case TYPE_BOOL:   set(x.b); break;
     case TYPE_NUMBER: set(x.n); break;
     case TYPE_TEXT:   set(x.t); break;
     case TYPE_FILE:   set(x.f); break;
     case TYPE_ARRAY:  set(x.a); break;
     case TYPE_DICT:  set(x.d); break;     
     case TYPE_PAIR:   set(x.p); break;
     default:          set(x.o);
                       break;
   }
 }
 
void dyn::set(std::vector<dyn> x) {
    flag = TYPE_ARRAY; 
    a = new solarray(x);
}

dyn::dyn(const dyn &x) {
  switch(x.flag) {
     case TYPE_BOOL:   set(x.b); break;
     case TYPE_NUMBER: set(x.n); break;
     case TYPE_TEXT:   set(x.t); break;
     case TYPE_FILE:   set(x.f); break;
     case TYPE_DICT:   set(x.d); break;
     case TYPE_ARRAY:  set(x.a); break;
     case TYPE_PAIR:   set(x.p); break;
     default:          set(x.o);       // object, tarray
                       break;
   }
 }

void dyn::set(dyn *x) {
  switch(x->flag)  {
     case TYPE_BOOL:   set(x->b); break;
     case TYPE_NUMBER: set(x->n); break;
     case TYPE_TEXT:   set(x->t); break;
     case TYPE_FILE:   set(x->f); break;
     case TYPE_DICT:   set(x->d); break;
     case TYPE_PAIR:   set(x->p); break;
     case TYPE_ARRAY:  set(x->a); break;
     default:          set(x->o);
                       break;
   }
 }

void dyn::set(const _arr32 &x) {
   _arr32 *arr = new _arr32((_arr32 *) &x);
   o = (void *) arr;
   flag = TYPE_ARRAY32;
}

void dyn::set(const _array64 &x)
{
   _array64 *arr = new _array64((_array64 *) &x);
   o = (void *) arr;
   flag = TYPE_ARRAY64;
}

void dyn::set(const _intArr &x)
{
   _intArr *arr = new _intArr((_intArr *) &x);
   o = (void *) arr;
   flag = TYPE_INTARR;
}


//===== convert, get


  text dyn::str(dyn d) { return d.toStr(); }

  bool     dyn::toBoolean(dyn d){ return d.toBoolean(); }
  int      dyn::toInt(dyn d)    { return d.toInt(); }
  natural  dyn::toNatural(dyn d) { return  d.toNatural(); }
  double   dyn::toNumber(dyn d) { return (double) d.toReal(); }
  double   dyn::toReal(dyn d)   { return (double) d.toReal(); }
  text     dyn::toText(dyn d)   { return d.toText(); }
  char     *dyn::toStr(dyn d)   { return d.toStr();}
  sol_pair *dyn::toPair(dyn d)  { return d.toPair(); }
  solarray    *dyn::toArray(dyn d) { return d.toArray(); }
  dict    *dyn::toDict(dyn d)   { return d.toDict(); }
  FILE     *dyn::toFile(dyn d)  { return d.toFile(); }
  void     *dyn::toList(dyn d)  { return d.toList(); }

bool dyn::toBoolean()
{
	if(flag == TYPE_BOOL) return b;
	if(flag == TYPE_NUMBER) return n != 0;	      
	if(flag == TYPE_TEXT) return t->compare("true") ? 1 : 0;
	return false;
}

int dyn::toInt()
{
	if(flag == TYPE_NUMBER) return (int) n;
	if(flag == TYPE_BOOL) return (int) b;	    
	if(flag == TYPE_TEXT) return (int) t->toNumber();
	return 0;
}

double  dyn::toReal() {
    if(flag == TYPE_NUMBER) return n;
    if(flag == TYPE_TEXT) return t->toNumber();
    return (double) 0;
  }

  double  dyn::toNumber() {
    if(flag == TYPE_NUMBER) return n;
    if(flag == TYPE_TEXT) return t->toNumber();
    return (double) 0;
  }


  natural dyn::toNatural() {
    if(flag == TYPE_NUMBER) return (natural) n;
    if(flag == TYPE_TEXT) return (natural) t->toNumber();
    return 0L;
 }

  solarray *dyn::toArray()
  {
    if(flag == TYPE_ARRAY) return a;
    return new solarray();
  }

  dict *dyn::toDict()  {
    if(flag == TYPE_DICT) return d;
    return new dict();
  }

  text dyn::toFunction()
  {
     return toText();
  }

  void *dyn::toObject()
  {
    if(flag == TYPE_OBJECT) return o;
    return (void *) NULL;
  }


  sol_pair *dyn::toPair()  { return p; }

static char *_sol_object = "Object";
static char _dyn_buffer[64];

char *dyn::toStr() {
  switch(flag) {
    case TYPE_NUMBER:
         sprintf(_dyn_buffer, "%g", n);
         break;
    case TYPE_TEXT:
         if(t != NULL) return t->data;
         return "";
    case TYPE_ARRAY:
         return a->toString();
    case TYPE_DICT:
         return d->toString();
    case TYPE_PAIR:
         return p->toStr();
    case TYPE_OBJECT:
         return _sol_object;
    case TYPE_BOOL:
         if(b) {
           strcpy(_dyn_buffer, "true");
           break;
         }
    default:
         strcpy(_dyn_buffer,"false");
         break;
   }
  return (char *) _dyn_buffer;
 }

text dyn::toText() {
   return text(toStr());
}

FILE *dyn::toFile()
{
  if(flag == TYPE_FILE) return f;
  return NULL;
}



/**
 * get the type of a typed solarray assigned to a dyn
 */

int dyn::arrayType()
{
  _arr32 *x = (_arr32 *) o;
  return x->type;
}

/**
 * toList(int = 0)
 * return the typed solarray
 * if the dyn is an solarray, convert to typed solarray according to the given type
 */

solarray *dyn::toList(int type) {
  solarray *arr = new solarray();
  switch(flag)  {
   case TYPE_NUMBER:
        arr->push(n);
        break;
   case TYPE_TEXT:
        arr->push(t);
        break;
   case TYPE_ARRAY:
        arr=a;
        break;
   case TYPE_DICT:
        arr = &d->values();
        break;
   default:  
        break;
  }
  return arr;
}

//===== utils

int dyn::len()
{
  switch(flag)
  {
   case TYPE_BOOL:    return 1;
   case TYPE_INT:     return 32;
   case TYPE_NATURAL:
   case TYPE_REAL:
   case TYPE_NUMBER:  return 64;
   case TYPE_TEXT:    return t->length();
   case TYPE_PAIR:    return 1;
   case TYPE_ARRAY:   return a->size();
   case TYPE_DICT:   return d->size();   
   case TYPE_ARRAY32: return ((_arr32 *)o)->size();
   case TYPE_ARRAY64: return ((_array64 *)o)->size();
//   case TYPE_DICT:  return di->size();
   default: break;
  }
  notYet();
  return 0;
}

int dyn::size() { return len(); }

void dyn::clone(dyn r)
{
  switch(flag)
   {
     case TYPE_BOOL:   r.set(b); break;
     case TYPE_NUMBER: r.set(n); break;
     case TYPE_TEXT:   r.set(t); break;
     case TYPE_PAIR:   r.set(p); break;
     case TYPE_ARRAY:  r.set(a); break;
     case TYPE_DICT:   r.set(d); break;
     default:          r.set(o); break;
   }
}

//======================================= Comparing

bool dyn::isNumber(const text &t)        { return text::isNumber(t.data); }
bool dyn::isNumber(text *t)       { return text::isNumber(t->data); }
bool dyn::isNumber(const char *x) { return text::isNumber(x); }
bool dyn::isNumber()
{
  switch(flag)
  {
    case TYPE_TEXT:   return isNumber(t);
    case TYPE_NUMBER: return true;
    default: break;
  }
  return false;
}

double dyn::toReal(const char *x)
{
  char *endptr;
  return strtod(x, &endptr);
}

bool dyn::equal(double d, const char *x)
{
  if(isNumber(x)) return d == toReal(x);
  return false;
}

bool dyn::equal(double d, text *t) { return equal(d, t->data); }

//=========================================== OPERATORS


void dyn::operator=(dyn d) { set(d); }


bool dyn::operator==(dyn *x) {
	switch(x->flag)	{
	case TYPE_BOOL:	return operator==(x->b);
	case TYPE_NUMBER:	return operator==(x->n);
	case TYPE_TEXT:	return operator==(x->t->data);
	case TYPE_FILE:		return f == x->f;
	case TYPE_ARRAY:	return a->operator==(x->a);
	case TYPE_DICT:   if(flag == TYPE_DICT) return d == x->d;
	default: 
		break;
	}
	return n == x->n;
}

bool dyn::operator==(const dyn &x)
{
  switch(x.flag)
  {
   case TYPE_BOOL:   return operator==(x.b);
   case TYPE_NUMBER: return operator==(x.n);
   case TYPE_TEXT:   return operator==(x.t->data);
   case TYPE_FILE:   return f == x.f;
   case TYPE_ARRAY:  return a->operator==(x.a);
   case TYPE_DICT:  return d->operator==(x.d);
   default: break;
  }
  return false;
}

bool dyn::operator==(int x)
{
  if(flag == TYPE_NUMBER) return n == (double) x;
  if(flag == TYPE_TEXT)   return equal((double) x, t->data);
  return false;
}

bool dyn::operator==(natural x)
{
  if(flag == TYPE_NUMBER) return n == (double) x;
  if(flag == TYPE_TEXT)   return equal((double) x, t->data);
  return false;
}

bool dyn::operator==(double x)
{
  if(flag == TYPE_NUMBER) return n == x;
  if(flag == TYPE_TEXT) return equal(x, t->data);
  return false;
}


bool dyn::operator==(bool x)
{
	switch(flag)
	{
	case TYPE_BOOL:   return b == x;
	case TYPE_TEXT:   return t->operator==(x);
	case TYPE_ARRAY:  return (a->size() != 0) == x;
	case TYPE_DICT:  return (d->size() != 0) == x;
	default: 
		break;
	}
	return (n != 0) == x;
}

bool dyn::operator==(const solarray &b)
{
  if(flag == TYPE_ARRAY) return a->operator==(b);
  return false;
}


bool dyn::operator==(const dict &b)
{
  if(flag == TYPE_DICT) return  d->operator==(b);
  return false;
}

bool dyn::operator==(const char *x)
{
  //cout << "dyn compare char * for << " << x << " type " << flag << endl;
  if(isNumber(x))
  {
    if(flag == TYPE_NUMBER) return n == toReal(x);
    if(flag == TYPE_TEXT)
    {
      if(isNumber(t)) return toReal(t->data) == toReal(x);
    }
  }
  if(flag == TYPE_TEXT) return t->identical(x) == 0;
  return false;
}

bool dyn::operator==(const text &x) {  return operator==(x.data); }

bool dyn::operator!=(int a)            { return ! operator==(a); }
bool dyn::operator!=(natural a)  { return ! operator==(a); }
bool dyn::operator!=(double a) { return ! operator==(a); }
bool dyn::operator!=(bool x)   { return ! operator==(x); }
bool dyn::operator!=(const dyn &a)    { return ! operator==(a); }
bool dyn::operator!=(dyn *a)   { return ! operator==(a); }
bool dyn::operator!=(const solarray &a)  { return ! operator==(a); }
bool dyn::operator!=(const char *a) { return ! operator==(a); }
bool dyn::operator!=(const text &a)   { return ! operator==(a); }
bool dyn::operator<(double d) {
  if(flag == TYPE_NUMBER) return n < d;
  if(flag == TYPE_TEXT)  {
    if(isNumber(t))  return toReal(t->data) < d;
  }
  return false;
}
bool dyn::operator<(int i) { return operator<((double) i); }
bool dyn::operator<(natural i) { return operator<((double) i); }
bool dyn::operator<(const char *x) {
  if(isNumber(x)) {
     if(flag == TYPE_NUMBER)  return n < toReal(x);
     if(flag == TYPE_TEXT) {
       if(isNumber(t)) return toReal(t->data) < toReal(x);
     }
  }

  if(flag == TYPE_TEXT) return t->identical(x) == -1;
  return false;
}
bool dyn::operator<(const text &x) { return operator<(x.data); }

bool dyn::operator<(bool x)
{
  if(flag == TYPE_BOOL) return b < x;
  return false;
}

bool dyn::operator<(const solarray &x)
{
  if(flag == TYPE_ARRAY) return a->operator<(x);
  return false;
}

bool dyn::operator<(dyn *x)
{
  switch(x->flag)
  {
    case TYPE_NUMBER: return operator<(x->n);
    case TYPE_TEXT:   return operator<(x->t->data);
    case TYPE_BOOL:   return operator<(x->b);
    case TYPE_ARRAY:  a->operator<(x->a);
    case TYPE_DICT:   return operator<(x->d);
    default: break;
  }
  return false;
}

bool dyn::operator<(const dyn &x)
{
  switch(x.flag)
  {
    case TYPE_NUMBER: return operator<(x.n);
    case TYPE_TEXT:   return operator<(x.t->data);
    case TYPE_BOOL:   return operator<(x.b);
    case TYPE_ARRAY:  a->operator<(x.a);
    case TYPE_DICT:   return operator<(x.d);
    default: break;
  }
  return false;
}

bool dyn::operator>(double d)
{
  if(flag == TYPE_NUMBER) return n > d;
  if(flag == TYPE_TEXT)
  {
    if(isNumber(t)) return t->toReal() > d;
  }
  return false;
}
bool dyn::operator>(int i) { return operator>((double) i); }
bool dyn::operator>(natural i) { return operator>((double) i); }
bool dyn::operator>(const char *x)
{
  if(isNumber(x))
  {
     if(flag == TYPE_NUMBER)  return n > toReal(x);
     if(flag == TYPE_TEXT)
     {
       if(isNumber(t)) return t->toReal() > toReal(x);
     }
  }

  if(flag == TYPE_TEXT) return t->identical(x) == 1;
  return false;
}

bool dyn::operator>(const text &x) { return operator>(x.data); }
bool dyn::operator>(bool x)
{
  if(flag == TYPE_BOOL) return b > x;
  return false;
}

bool dyn::operator>(const solarray &x) {
  if(flag == TYPE_ARRAY) return a->operator>(x);
  return false;
}

bool dyn::operator>(const dict &x) {
  if(flag == TYPE_DICT) return d->operator>(x);
  return false;
}


bool dyn::operator>(dyn *x)
{
  switch(x->flag)
  {
    case TYPE_NUMBER: return operator>(x->n);
    case TYPE_TEXT:   return operator>(x->t->data);
    case TYPE_BOOL:   return operator>(x->b);
    case TYPE_ARRAY:  return a->operator>(x->a);
    case TYPE_DICT:   return d->operator>(x->d);
    default: break;
  }
  return false;
}

bool dyn::operator>(const dyn &x)
{
  switch(x.flag)
  {
    case TYPE_NUMBER: return operator>(x.n);
    case TYPE_TEXT:   return operator>(x.t->data);
    case TYPE_BOOL:   return operator>(x.b);
    case TYPE_ARRAY:  return a->operator>(x.a);
    case TYPE_DICT:   return d->operator>(x.d);
    default: break;
  }
  return false;
}

//==================================== ARITHMETIC

#define BOOL_TO_NUM if(flag == TYPE_BOOL) boolToNum();

void dyn::boolToNum()
{
   (b == true) ? n = 1 : n = 0;
   flag = TYPE_NUMBER;
}

double dyn::operator+(int a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n + a;
  nomatch('+');
  return (double) 0;
}

double dyn::operator+(natural a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n + (double) a;
  nomatch('+');
  return (double) 0;
}

double dyn::operator+(double a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n + a;
  nomatch('+');
  return (double) 0;
}

char *dyn::operator+(const char *x)
{
  if(flag == TYPE_TEXT) {
   text *y = new text(t);
   y->cat(x);
   return y->data;
  }
  nomatch('+');
  return "";
}


solarray dyn::operator+(const solarray &x)
{
  if(flag == TYPE_ARRAY) return a->operator+(x);
  nomatch('+');
  return solarray();
}

solarray dyn::operator+(solarray *x)
{
  if(flag == TYPE_ARRAY) return a->operator+(x);
  nomatch('+');
  return solarray();
}

text dyn::operator+(const text &x) { return text(operator+(x.data)); }

dyn dyn::operator+(const dyn &x) {
  switch(x.flag)  {
   case TYPE_BOOL:
        if(x.b == true) return operator+(1);
        return this;
   case TYPE_NUMBER:
        BOOL_TO_NUM
        return operator+(x.n);
   case TYPE_TEXT:
        return t->operator+(x.t);
    case TYPE_ARRAY:    
        if(flag == TYPE_ARRAY) return a->operator+(x.a);
        a->push(x);
        return this;   
    case TYPE_DICT:
        if(flag == TYPE_DICT) return d->operator+(x.d);
        d->push(x);
        return this;        
   default:
        break;
  }
  nomatch('+');
  return (double) 0;
}

double dyn::operator-(int a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n - a;
  nomatch('-');
  return (double) 0;
}

double dyn::operator-(natural a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n - (double) a;
  nomatch('-');
  return (double) 0;
}


double dyn::operator-(double a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n - a;
  nomatch('-');
  return (double) 0;
}

dyn dyn::operator-(const dyn &x)
{
  switch(x.flag)
  {
   case TYPE_BOOL:
        if(x.b == true) return operator-(1);
        return dyn(this);
   case TYPE_NUMBER:
        BOOL_TO_NUM
        return operator-(x.n);
   case TYPE_TEXT:
        return operator-(x.t->toNumber());
   case TYPE_ARRAY:
        return a->operator-(x.a);
   case TYPE_DICT:
        return d->operator-(x.d);        
   default:
        break;
  }
  nomatch('-');
  return (double) 0;
}

solarray dyn::operator-(const solarray &b)
{
  if(flag == TYPE_ARRAY)  return a->operator-(b);
  nomatch('-');
  return solarray();
}


double dyn::operator*(int a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n * a;
  nomatch('*');
  return (double) 0;
}

double dyn::operator*(natural a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n * (double) a;
  nomatch('*');
  return (double) 0;
}


double dyn::operator*(double a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n * a;
  nomatch('*');
  return (double) 0;
}
double dyn::operator*(dyn x)
{
  switch(x.flag)
  {
   case TYPE_BOOL:   (x.b == true) ? x.n = 1 : x.n = 0;
   case TYPE_NUMBER: BOOL_TO_NUM
                     return operator*(x.n);
   default: break;
  }
  nomatch('*');
  return (double) 0;
}


double dyn::operator%(int a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return fmod(n , (double) a);
  nomatch('%');
  return (double) 0;
}

double dyn::operator%(natural a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return fmod(n , (double) a);
  nomatch('%');
  return (double) 0;
}


double dyn::operator%(double a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return fmod(n , a);
  nomatch('%');
  return (double) 0;
}
double dyn::operator%(dyn x)
{
  switch(x.flag)
  {
   case TYPE_BOOL:   (x.b == true) ? x.n = 1 : x.n = 0;
   case TYPE_NUMBER: BOOL_TO_NUM
                     return operator%(x.n);
   default: break;
  }
  nomatch('%');
  return (double) 0;
}

double dyn::operator/(int a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n / a;
  nomatch('/');
  return (double) 0;
}

double dyn::operator/(natural a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n / (double) a;
  nomatch('/');
  return (double) 0;
}


double dyn::operator/(double a)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) return n / a;
  nomatch('/');
  return (double) 0;
}

double dyn::operator/(dyn x)
{
  switch(x.flag)
  {
   case TYPE_BOOL:   (x.b == true) ? x.n = 1 : x.n = 0;
   case TYPE_NUMBER: BOOL_TO_NUM
                     return operator/(x.n);
   default: break;
  }
  nomatch('/');
  return (double) 0;
}

int dyn::operator&(int i)
{
  if(flag == TYPE_NUMBER) return (int) n & i;
  return 0;
}

int dyn::operator&(double i)
{
  if(flag == TYPE_NUMBER) return (int) n & (int) i;
  return 0;
}

int dyn::operator&(dyn i)
{
  if(flag == TYPE_NUMBER && i.flag == TYPE_NUMBER) return (int) n & (int) i.n;
  return 0;
}


int dyn::operator|(int i)
{
  if(flag == TYPE_NUMBER) return (int) n | i;
  return i;
}

int dyn::operator|(double i)
{
  if(flag == TYPE_NUMBER) return (int) n | (int) i;
  return (int) i;
}

int dyn::operator|(dyn i)
{
  if(flag == TYPE_NUMBER && i.flag == TYPE_NUMBER) return (int) n | (int) i.n;
  return (int) i.n;
}


bool dyn::operator&&(dyn d)
{
  return toBoolean() && d.toBoolean();
}
bool dyn::operator&&(int d)
{
  return toBoolean() && d;
}

bool dyn::operator&&(double d)
{
  return toBoolean() && (int) d;
}


bool dyn::operator||(dyn d)
{
  return toBoolean() || d.toBoolean();
}

bool dyn::operator||(int d)
{
  return toBoolean() || d;
}

bool dyn::operator||(double d)
{
  return toBoolean() || (int) d;
}



//================================= Augmented

void dyn::operator+=(dyn x)
 {
   switch(x.flag)
   {
    case TYPE_BOOL:
         if(x.b == true) operator+=(1);
         break;
    case TYPE_NUMBER:
         BOOL_TO_NUM
         operator+=(x.n);
         return;
    case TYPE_TEXT:
         operator+=(x.t->toNumber());
         return;
    case TYPE_ARRAY:
         a->operator+=(x);
         return;
    case TYPE_DICT:
         d->operator+=(x);
         return;         
    default: break;
   }
   nomatch('>');
 }

void dyn::operator+=(int a) { operator+=((double) a);}
void dyn::operator+=(natural a) { operator+=((double) a);}
void dyn::operator+=(double a)
 {
   BOOL_TO_NUM
   if(flag == TYPE_NUMBER) { set(n + a); return;}
   if(flag == TYPE_TEXT)
   {
     if(t->isNumber()) { set(t->toNumber() + a); return; }
   }
   nomatch('>');
 }

void dyn::operator+=(const text &b)
{
  if(flag == TYPE_TEXT) { t->cat(b.data); return; }
  nomatch('>');
}
void dyn::operator+=(const char *b) { operator+=(text(b)); }

void dyn::operator+=(const solarray &x)
{
  if(flag == TYPE_ARRAY) { a->operator+=(x); return; }
  nomatch('>');
}


void dyn::operator-=(double x)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) { set(n - x); return; }
  if(flag == TYPE_TEXT)
   {
     if(t->isNumber()) { set(t->toNumber() - x); return; }
   }
  nomatch('<');
}

void dyn::operator-=(int a) { operator-=((double) a);}
void dyn::operator-=(natural a) { operator-=((double) a);}

void dyn::operator-=(dyn x)
 {
   switch(x.flag)
   {
      case TYPE_BOOL:
           if(x.b == true) operator-=(1);
           break;
      case TYPE_NUMBER:
           BOOL_TO_NUM
           operator-=(x.n);
           break;
//      case TYPE_TEXT:   operator-=(x.t->toNumber()); break;
      case TYPE_ARRAY:    a->operator-=(x); break;
      default: nomatch('<');
   }
 }

void dyn::operator-=(const solarray &x)
{
  if(flag == TYPE_ARRAY) { a->operator-=(x); return; }
  nomatch('<');
}


void dyn::operator*=(int x) { operator*=((double) x); }
void dyn::operator*=(natural x) { operator*=((double) x); }
void dyn::operator*=(double x)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) { set(n * x); return; }
  if(flag == TYPE_TEXT)
   {
     if(t->isNumber()) { set(t->toNumber() * x); return; }
   }
  nomatch('S');
}
void dyn::operator*=(dyn x)
{
   switch(x.flag)
   {
      case TYPE_BOOL:
           if(x.b != true) operator*=(0);
           break;
      case TYPE_NUMBER:
           BOOL_TO_NUM
           operator*=(x.n);
           break;
      case TYPE_TEXT:
           operator*=(x.t->toNumber());
           break;
      default: nomatch('S');
   }

}

void dyn::operator/=(int x) { operator/=((double) x); }
void dyn::operator/=(natural x) { operator/=((double) x); }

void dyn::operator/=(double x)
{
  BOOL_TO_NUM
  if(flag == TYPE_NUMBER) { set(n / x); return; }
  if(flag == TYPE_TEXT)
   {
     if(t->isNumber()) { set(t->toNumber() / x); return; }
   }
  nomatch('D');
}

void dyn::operator/=(dyn x)
{
  switch(x.flag)
  {
    case TYPE_BOOL:
         if(x.b != true) operator/=(0);
         break;
    case TYPE_NUMBER:
         BOOL_TO_NUM
         operator/=(x.n);
         break;
    case TYPE_TEXT:
         operator/=(x.t->toNumber());
         break;
    default: nomatch('D');
  }
}


bool dyn::operator!()
{
  switch(flag)
  {
    case TYPE_BOOL:    return !b;
    case TYPE_INT:
    case TYPE_NUMBER:  return n == 0;
    case TYPE_TEXT:    return t->length() == 0;
    case TYPE_ARRAY:   return a->size() == 0;
    case TYPE_DICT:   return d->size() == 0;    
    case TYPE_FILE:    return f == NULL;
    default: nomatch('!'); break;
  }
  return false;
}

solarray dummyArray = solarray();
solarray &dyn::operator,(dyn x)
{
  if(flag == TYPE_ARRAY)
  {
    a->push(x);
    return *a;
  }
  nomatch(',');
  return dummyArray;
}

void dyn::operator<<=(int x)
{
  if(flag == TYPE_INT) n = (double) (((int) n) << x);
  else nomatch('<');
}
void dyn::operator>>=(int x)
{
  if(flag == TYPE_INT) n = (double) (((int) n) >> x);
  else nomatch('>');
}


//======================================= text

#define ISTEXT if(flag != TYPE_TEXT) return "";

void dyn::cat(dyn x)   { t->cat(x.t->data); }
void dyn::cat(char *x) { t->cat(x); }
void dyn::cat(const text &x)  { cat(x.data); }

text dyn::trim()
{
  ISTEXT
  return t->trim();
}
text dyn::lTrim()
{
  ISTEXT
  return t->lTrim();
}
text dyn::rTrim()
{
  ISTEXT
  return t->rTrim();
}
text dyn::lower()
{
  ISTEXT
  return t->lower();
}
text dyn::upper()
{
  ISTEXT
  return t->upper();
}

//===== solarray

void dyn::display() {
  _arr32 *x;
  _array64 *y;

  switch(flag)
  {
   case TYPE_ARRAY:
        a->display();
        return;
   case TYPE_DICT:
        d->display();
        return;        
   case TYPE_ARRAY32:
        x = (_arr32 *) o;
        x->display();
        return;
   case TYPE_ARRAY64:
        y = (_array64 *) o;
        y->display();
        return;
   default:
        break;
  }
   std::cout << toStr() << std::endl;
 }

void dyn::reset()      { a->reset(); }
dyn  dyn::inc()        { return a->inc(); }
dyn  dyn::dec()        { return a->dec(); }
dyn  dyn::value()      { return a->value(); }
dyn  dyn::key()        { return a->key(); }
dyn  dyn::begin()      { return a->begin(); }
dyn  dyn::end()        { return a->end(); }
int  dyn::index()      { return d->index(); }

//============ stack

void dyn::push(dyn b)
{
 if (flag == TYPE_ARRAY) a->push(b);
 else nomatch('U');
}

dyn dyn::pop()
{
  if (flag == TYPE_ARRAY) return a->pop();
  nomatch('O');
  return dyn(false);
}

dyn dyn::shift() { return a->shift(); }
void dyn::unshift(dyn x) { a->unshift(x); }

solarray dyn::slice(int s, int e, int i) { return a->slice(s, e, i); }
void dyn::splice(int s, int e, solarray x, int i) { a->splice(s, e, x, i); }

solarray dyn::slice(natural s, natural e, natural i)
 { return a->slice((int) s, (int) e, (int) i); }
void dyn::splice(natural s, natural e, solarray x, natural i)
 { a->splice((int) s, (int) e, x, (int) i); }

double dyn::sum() { return a->sum(); }

//=========================================== display

ostream& operator<< (ostream& out, dyn d)
{
  _arr32 *x;
  _array64 *y;

  switch(d.flag)
  {
   case TYPE_BOOL:    if(d.b == 0) return out  << "false";
                      return out << "true";
   case TYPE_TEXT:    return out << d.t->data;
   case TYPE_NUMBER:  return out << d.n;
   case TYPE_ARRAY:   return out << d.a->toText().data;
   case TYPE_DICT:   return out << d.d->toText().data;   
   case TYPE_ARRAY32: x = (_arr32 *) d.o;
                      return out << x->toText();
   case TYPE_ARRAY64: y = (_array64 *) d.o;
                      return out << y->toText();
   case TYPE_OBJECT:  return out << "Object";
   default: break;
 }
 unknown("<<");
 return out;
}

ostream& operator<< (ostream& out, dyn *d)
{
  switch(d->flag)
  {
   case TYPE_BOOL:   if(d->b == 0) return out  << "false";
                     return out << "true";
   case TYPE_TEXT:   return out << d->t->data;
   case TYPE_NUMBER: return out << d->n;
   case TYPE_ARRAY:  return out << d->a->toText().data;
   case TYPE_DICT:  return out << d->d->toText().data;   
   default: break;
 }
 unknown("<<");
 return out;
}

void dyn::setAt(int pos, const char *s) { setAt(pos, text(s)); }
void dyn::setAt(int pos, const text &x)
{
  if(flag == TYPE_TEXT) { t->setAt(pos, x); return; }
  setAt(pos, dyn(x));
}

void dyn::setAt(int pos, dyn x)
{
 if(flag == TYPE_ARRAY) { a[pos] = x; return; }
 nomatch('A');
}

dyn dyn::operator[](int i)
{
  if(flag == TYPE_ARRAY) return a->operator[](i);
  else
   if(flag == TYPE_TEXT)  return dyn(t->operator[](i));
   else
   {
    nomatch('[');
  }
  return dyn(false);
}
dyn  dyn::operator[](const text &t) { return operator[](t.data); }
dyn  dyn::operator[](const char * s)
{
  if(flag == TYPE_DICT) return d->operator[](s);
  nomatch('[');
  return dyn(false);
}
dyn  dyn::operator[](dyn d)
{
  if(d.flag == TYPE_NUMBER) return operator[]((int) n);
  if(d.flag == TYPE_TEXT)   return operator[](t->data);
  nomatch('[');
  return dyn(false);
}

dyn dyn::at(int i)
{
  if(flag == TYPE_ARRAY) return a->at(i);
  if(flag == TYPE_TEXT) return t[i];
  nomatch('A');
  return dyn(false);
}


dyn dyn::at(const text &t)
{
  if(flag == TYPE_DICT) return d->at(t);  
  nomatch('A');
  return dyn(false);
}
dyn dyn::at(const char *s)
{
  if(flag == TYPE_DICT) return d->at(text(s));
  nomatch('A');
  return dyn(false);
}


text operator+(text t, dyn d)        { return t.operator+(d.toText()); }
text operator+(const char *t, dyn d) { return text(t).operator+(d.toText()); }
dyn operator+(int a, dyn b)          { return b.operator+(a); }
dyn operator+(double a, dyn b)       { return b.operator+(a); }
dyn operator+(solarray *a, dyn b)       { return b.operator+(a); }

dyn operator-(int a, dyn b)       { return dyn(a) - b; }
dyn operator-(double a, dyn b)    { return dyn(a) - b; }
dyn operator-(solarray *a, dyn b)    { return dyn(a) - b; }

dyn operator*(int a, dyn b)       { return b.operator*(a); }
dyn operator*(double a, dyn b)    { return b.operator*(a); }

dyn operator/(int a, dyn b)    { return dyn(a) / b; }
dyn operator/(double a, dyn b) { return dyn(a) / b; }

int operator&(int a, dyn b)    { return  b & a; }
int operator&(double a, dyn b) { return  b & a; }

int operator|(int a, dyn b)    { return  b | a; }
int operator|(double a, dyn b) { return  b | a; }

void operator+=(int a, dyn b)    { a = a + b.toInteger(); }
void operator+=(double a, dyn b) { a = a + b.toReal(); }
void operator+=(text t, dyn d)   { t = t + d.toText(); }
void operator+=(const char *t, dyn d)  { t = text(t).operator+(d.toText()).toStr(); }

void operator-=(int a, dyn b)    { a = a - b.toInteger(); }
void operator-=(double a, dyn b) { a = a - b.toReal(); }

void operator*=(int a, dyn b)    { a = b.toInteger() * a; }
void operator*=(double a, dyn b) { a = b.toReal() * a; }

void operator/=(int a, dyn b)    { a = a / b.toInteger(); }
void operator/=(double a, dyn b) { a = a / b.toReal(); }

void operator&=(int a, dyn b)    { a = b.toInteger() & a; }
void operator&=(double a, dyn b) { a = (double) (b.toInteger() & (int) a); }

void operator|=(int a, dyn b)    { a = b.toInteger() | a; }
void operator|=(double a, dyn b) { a = (double) (b.toInteger() | (int) a); }

bool operator&&(int a, dyn b)    { return b.operator&&(a); }
bool operator&&(double a, dyn b) { return b.operator&&(a); }

bool operator||(int a, dyn b)    { return b.operator||(a); }
bool operator||(double a, dyn b) { return b.operator||(a); }






