/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>
#include <math.h>
#include <stdarg.h>
#include <dirent.h>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include <iostream>
#include <vector>

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "fun.hpp"



//===== general

void runtimeError(const char *mess) { printf("%s\n", mess); }

/**
  Get the minimum of two values
  */
int min(int a, int b) { return a <= b ? a : b; }
natural min(natural a, natural b) { return a <= b ? a : b; }
double min(double a, double b) { return a <= b ? a : b; }
dyn min(dyn a, dyn b) { return a <= b ? a : b; }
text min(text a, text b) { return a <= b ? a : b; }

/**
  Get the maximum of two values
  */
int max(int a, int b) { return a > b ? a : b; }
natural max(natural a, natural b) { return a > b ? a : b; }
double max(double a, double b) { return a > b ? a : b; }
dyn max(dyn a, dyn b) { return a > b ? a : b; }
text max(text a, text b) { return a > b ? a : b; }

/* Sign */

int _sol_sgn(double x) {
    if(x == 0) return 0;
    return (x < 0) ? -1 : 1;
};

int _sol_sgn(int x) {
    if(x == 0) return 0;
    return (x < 0) ? -1 : 1;
};



/**
  Displays a text and stops the program.
  */

void die(text t) { die(t.data); }

/**
 Display a string and stops the program.
 */
void die(const char *s) { printf("%s\n", s); exit(1); }

/**
  Calls the exit(0) system command.
  */

void exit() { exit(0); }


/**
  Make the plural of a name,
  displays "s" if parameter's value is 2 or more.
  */

const char *plural(int x) { return( x > 1 ? "s" : "" ); }

/**
  Converts a dyn to natural.
  */

natural naturalval(dyn d) { return d.toNatural(); }

/**
  Converts a char to natural.
  */
natural naturalval(unsigned char x) { return (natural) x; }

/**
  Converts a 32 bits signed integer to natural.
  */

natural naturalval(int x) { return (natural) x; }
natural naturalval(natural x) { return  x; }

/**
  Converts a double to natural.
  */

natural naturalval(double x) { return (natural) x; }

/**
  Converts the text representation of a number to natural.
  */

natural naturalval(text x) { return x.toNatural(); }

/**
  Converts the string representation of a number to natural.
  */

natural naturalval(const char *s) {
 char *endptr;
 return (natural) strtod(s, &endptr);
}

natural naturalval(void *x) { return (natural)(int) x; }


bool boolval(int i) { return i != 0; }

//===== types

/**
  Returns true is the dyn is an associative array.
  */

bool is_dyn_array(dyn a)  { return a.isArray();  }
bool is_dyn_dict(dyn a)  { return a.isDict();  }
bool is_dyn_object(dyn a)  { return a.isObject();  }

/**
  Returns true is the dyn is a typed array.
  Look at the arrayType() method of dyn for the type.
  */

bool is_dyn_list(dyn d)  { return d.isList(); }

//===== numbers

double doubleval(text s) { return doubleval(s.data); }
double doubleval(const char *s) { char *endptr; return strtod(s, &endptr); }
double doubleval(dyn d) { return d.toReal(); }
double doubleval(int i) { return (double) i; }
double doubleval(natural i) { return (double) i; }
double doubleval(double d) { return d; }
double doubleval(void * i) { return (double)(int) i; }

int intval(text s) { return atoi(s.data); }
int intval(const char *s) { return atoi(s); }
int intval(dyn d) { return d.toInt();  }
int intval(double d) { return (int) d; }
int intval(int i) { return i; }
int intval(natural i) { return (int) i; }
int intval(void *v) { return (int) v; }

text strval(int i)     { return text(i); }
text strval(natural i) { return text(i); }
text strval(double d)  { return text(d); }
text strval(dyn d)     { return d.toText(); }
text strval(void *v)   { return text((char *) v); }
text strval(bool b)    { return (b == true) ? text("true") : text("false"); }
text strval(text t)    { return t; }
text strval(char *s)   { return text(s); }


/**
  Returns the integer value represented by the string.
  */

int ord(const char *s) {
     if(strlen(s) > 0) return (int) s[0];
     else return 0;
}

/**
  Returns the integer value represented by the text.
  */

int ord(text t) { return ord(t.data); }
int ord(char c) { return (int) c; }

/**
  Converts an integer to a text.
  */

text chr(int i) 
{
	char c[2];
	c[0] = (char) (i & 255);
	c[1]='\0';
	return text(c);
}

double _sol_round(double x) { return floor(x + 0.5); }


//===== TEXT


char builtin_buffer[32];

/**
  Make a text from a boolean
  */

char *str(bool b)
{
  char *buffer = (char *) new char[16];
  sprintf(buffer, "%d", (int) b);
  return buffer;
}

/**
  Make a text representing an integer.
  */

char *str(int i)
{
  char *buffer = (char *) new char[16];
  sprintf(buffer, "%d", i);
  return buffer;
}

/**
  Make a text representing a number.
  */

char *str(double d)
{
  char *buffer = (char *) new char[32];
  sprintf(buffer, "%g", d);
  return buffer;
}

/**
  Make a text from a dyn holding a scalar.
  */

char *str(dyn a)    { return a.toStr(); }

/**
  Make a text from the object pointed out.
  */

char *str(void *s)  { return (char *) s; }


char *str(text t)  { return (t.data); }
char *str(char *s)  { return (s); }


/* return an integer as an hexa text: 0x00000000 */
char *builtin_hex(int i)
{
  sprintf(builtin_buffer, "%x", i);
  return builtin_buffer;
}


char *strtolower(const char *a)
{
  int i = strlen(a);
  char *b = (char *) new char[i+1]; 
  while(--i >= 0) b[i] = (char) tolower(a[i]);
  return b;
}

char *strtoupper(const char *a)
{
  int i = strlen(a);
  char *b = (char *) new char[i+1];
  while(--i >= 0) b[i] = (char) toupper(a[i]);
  return b;
}

char *ucfirst(const char *s)
{
  int l = strlen(s);
  char *x = (char *) new char[l+1];
  strcpy(x, s);
  x[0] = (char) toupper(x[0]);
  return x;
}

char *ucwords(const char *s)
{
  bool flag = true;
  char c;
  int l = strlen(s);
  char *x = (char *) new char[l+1];

  for(int i = 0; i < l; i++)
  {
    c = s[0];
    if(c > 32)
    {
     if(flag) c  = (char) toupper(c);
     flag = false;
    }
    else flag = true;
    x[0] = c;
  }
  return x;
}

// from pos for l chars

char *substr(const char *s, int pos) { return substr(s, pos, strlen(s)); }
char *substr(const char *s, int pos, int l)
{
  char *nstr = new char[l + 1];
  strncpy(nstr, s + pos, l);
  nstr[l] = '\0';
  return nstr;
}



char *chop(const char *s) { return rtrim(s); }

char *rtrim(const char *s)
{
  int l = strlen(s);

  char c;
  int i;

  for (i = l - 1; i >= 0; i--)
  {
    c = s[i];
    if (c <= (char) 32) l--;
    else break;
  }
  return substr(s, 0, l);
}

char *ltrim(const char *s)
{
  int l = strlen(s);
  char c;
  int i;

  for (i = 0; i < l; i++)
  {
    c = s[i];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') continue;
    else break;
  }
  return substr(s, i, l - i);
}

char *trim(const char *s)
{
  char c;
  int l, i, j;
  int len = strlen(s);

  for (i = 0; i < len; i++)
  {
    c = s[i];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') continue;
    else break;
  }
  l = len - i;
  for (j = len - 1; j >= i; j--)
  {
    c = s[j];
    if (c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '\v') l--;
    else  break;
  }
  return substr(s, i, l);
}

#ifdef _UNIX
#ifdef VISUALC

int strcasecmp(const char *a, const char *b)
{
// return strcmp(strtolower(a), strtolower(b));
  char ca;
  char cb;
  while(true)
  {
   ca = tolower(*a++);
   cb = tolower(*b++);
   if(ca < cb) return -1;
   if(ca > cb) return 1;
   if(ca == 0) return 0;
  }
}

#endif
#endif

char *stristr(const char *s, const char *n)
{
  int i = text(s).findLex(n);
  if(i == -1) return NULL;
  return substr(s, i);
}


int strpos(const char *s, const char *sea, int stt)
{
  int nl = strlen(sea);
  int l = strlen(s) - nl + 1;
  for(int i = stt; i < l; i++)
  {
    //cout << "compare " << (s + i) << " x [" << sea << "]" << endl;
    if(strncmp(s + i, sea, nl) == 0) return i;
  }
  return -1;
}

char *str_replace(const char *str, const char *s, const char *r)
{
  text t = text(str);
  t.replace(s, r);
  return t.data;
}

char *strtr(const char *str, const char *sea, const char *rep)
{
  char c;
  int i, j;
  int l = strlen(str);
  int ls = strlen(sea);
  int l2 = strlen(rep);  if(l2 < ls) ls = l2;
  char *n = new char[l + 1];
  for(i = 0; i < l; i++)
  {
    c = str[i];
    for(j = 0; j < ls; j++)
    {
     if(c == sea[j]) { c = rep[j]; break; }
    }
    n[i] = c;
  }
  n[l] = '\0';
  return n;
}

int substr_count(char *str, const char *f)
{
  int count =  -1;
  do
  {
    count++;
    str = strstr(str, f);
  } while(str != NULL);
  return count;
}

char *substr_replace(const char *str, const char *rep, int a, int b)
{
  text s = substr(str, 0, a);
  s.cat(rep);
  s.cat(substr(str, b + 1, strlen(str)));
  return s.data;
}

text str_pad(text t, int n, text padding, int mode)
 { return str_pad(t.data, n, padding.data, mode); }
text str_pad(text t, int n, const char *padding, int mode)
 { return str_pad(t.data, n, padding, mode); }
text str_pad(const char *t, int newsize, const char *padding, int mode)
{
  int plen = strlen(padding);
  int olen = strlen(t);
  int size = newsize;
  size -= olen;
  if(size < 1) return text(t);

  if(mode == STR_PAD_BOTH) size = (size + 1) / 2;
  char *s = (char *) new char[size+1];
  int off = 0;

  //cout << "str_pad " << t << " with [" << padding << "] size=" << olen << " add " << size << endl;

 loop:
  {
    if((off + plen) <= size)
    {
       strncpy(s + off, padding, plen);
    }
    else
    {
       plen = size - off;
       strncpy(s + off, substr(padding, 0, plen), plen);
    }
    off += plen;
    if(off < size) goto loop;
  }
  s[size] = '\0';

  text x = "";
  if(mode == STR_PAD_LEFT)  x.set(s);
  if(mode == STR_PAD_BOTH)
  {
    x.set(s);
    if((size + size + olen) > newsize) x = x.slice(0, size - 1, 0);
  }
  x.cat(t);
  if(mode == STR_PAD_RIGHT || mode == STR_PAD_BOTH)  x.cat(s);
  if(x.length() > newsize) return x.slice(0, newsize, 0);
  return x;
}


char *str_repeat(const char *src, int mult)
{
  return (char *) text(src).dup(mult).c_str();
}

//===== regular not yet implemented

/*
bool ereg(text t, text sea) { return ereg(t.data, sea.data); }
bool ereg(text t, const char *sea)  { return ereg(t.data, sea); }
bool ereg(const char *t, text sea)  { return ereg(t, sea.data); }
bool ereg(const char *t, const char *sea) { return false; }
*/



//===============


int time()
{
  time_t t = time(NULL);
  return (int) t;
}

char *input(const char *message)
{
    printf("%s", message);
    char *temp = (char *) new char[256];
    std::cin.getline(temp, 256, '\n');
    return temp;
}

	
char *_sol_inputText(const char *message)
{
	char *str;
	char buffer[320];
	
	printf("%s ", message);
	fgets( buffer, 255, stdin );
	int len = strlen((char *) buffer);
	
	str = (char *) new char[len+1];
	strcpy(str,  (char *) buffer); 
	return str;  
}

double _sol_inputNumber(const char *message)
{
	char *str;
	char buffer[320];
	
	printf("%s ", message);	
	fgets( buffer, 255, stdin );
	int len = strlen((char *) buffer);
	
	str = (char *) new char[len+1];
	strcpy(str,  (char *) buffer); 
  
	return atof(str);

}

/**
 * Sol PHP string
 * 
 * This function accept a variable list of string and return
 * them as a  concatenated string
 */    

char *_sol_phpstring(const char *fmt, ... )
 {
    char *value;
    char *buffer;
    
    va_list thelist;
    
    va_start(thelist, fmt );    
    int size = 0;

looplen:
	value = va_arg(thelist, char * );

	if(value != NULL)
	{
		size += strlen(value);
		goto looplen;
	}	
    va_end( thelist );
	
	buffer = (char *) new char[size+1];
	buffer[0] = '\0';
	
    va_start(thelist, fmt );	

loopcat:
	value = va_arg(thelist, char * );

	if(value != NULL)
	{
		buffer = strcat(buffer, value);
		goto loopcat;
	}	
    va_end( thelist );

	return buffer;
 }
 
 char *_sol_phpcvt(text *t)
 {
 	return t->data;
 }

 char *_sol_phpcvt(dyn *d)
 {
 	return d->toStr();
 }

/* dump solarray */

char * _sol_dumpArray(solarray a)
{
  return a.toString();
}
/*
char * _sol_dumpArray(_array32 a)
{
  return a.toString();
}
char * _sol_dumpArray(_array64 a)
{
  return a.toString();
}
*/

