/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>
#include <math.h>
#include <dirent.h>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include <iostream>
#include <vector>
#include <map>

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "dict.hpp"
#include "fun.hpp"

//===== solarray

int count(solarray a) { return a.size(); }

solarray arrayval(_arr32 ta) {
    solarray s = solarray();
    for(int i = 0; i < ta.len; i++) {
        s.push(dyn(ta[i]));
    }
    return s;
}

solarray arrayval(_intArr ia) {
    solarray s = solarray();
    for(int i = 0; i < ia.len; i++) {
        s.push(dyn(ia[i]));
    }
    return s;
}

solarray arrayval(_array64 ra) {
    solarray s = solarray();
    int type = ra.type;
    for(int i = 0; i < ra.len; i++) {
        dyn x = dyn();
        x.set(*ra[i]);
        s.push(x);
    }
    return s;
}


solarray  _sol_range(int a, int b) {
  solarray x = solarray();

  if(a > b)  {
    while(a > b) {
     x.push(dyn(a));
     a--;
    }
  }
  else {
    while(a <= b)  {
     x.push(dyn(a));
     a++;
    }
  }
  return x;
}

solarray array_merge(solarray a, solarray b) { return a + b; }
solarray array_diff(solarray a, solarray b) { return a - b; }
solarray array_intersect(solarray a, solarray b) { return a & b; }
solarray array_unique(const solarray &a) { solarray b = a; b.unique(); return b; }


bool in_array(int i, solarray a)    { return a.in(new dyn(i)); }
bool in_array(text s, solarray a) { return a.in(new dyn(s)); }
bool in_array(const char s, solarray a) { return a.in(new dyn(s)); }
bool in_array(double d, solarray a) { return a.in(new dyn(d)); }
bool in_array(dyn d, solarray a)    { return a.in(d); }

// for control structure using either text or solarray
bool in_sequence(const char *x, text &t) { return t.in(x); }
bool in_sequence(const char *x, const char *t) { return text(t).in(x); }


void array_map(void(*fun)(dyn &), solarray &a)
{
  a.begin();
  for(int i = 0; i < a.size(); i++)
   {
     fun(a[i]);
     a.inc();
   }
}

void array_map(void(*fun)(dyn *), solarray &a) {
  a.begin();
  for(int i = 0; i < a.size(); i++) {
     fun(&a[i]);
     a.inc();
   }
}


void array_map(void (*fun)(dyn &, dyn &), solarray &a, solarray &b) {
  int i = 0;
  while(i < a.size() && i < b.size())  {
    if(i == 0) { a.begin(); b.begin();  }
    else      { a.inc(); b.inc(); }
    fun(a[i], b[i]);
    i++;
  }
}

void array_map(void (*fun)(dyn &a, dyn &b, dyn &c), solarray &a, solarray &b, solarray &c) {
  int i = 0;
  while(i < a.size() && i < b.size() && i < c.size())  {
    if(i == 0) { a.begin(); b.begin(); c.begin();  }
    else      { a.inc(); b.inc(); c.inc(); }
    fun(a[i], b[i], c[i]);
    i++;
  }
}


//===== iterators

dyn reset(solarray &x) { return x.begin(); }
dyn next(solarray &x)  { return x.inc(); }
dyn prev(solarray &x)  { return x.dec(); }
dyn end(solarray &x)   { return x.end(); }
text key(solarray &x)  { return x.key(); }
dyn current(solarray &x) { return x.value(); }

//===== functions on arrays

text implode(solarray a, const char *s) { return a.join(s); }

solarray explode(const char *s, const char *sep)
{
  solarray x;
  const char *nstr;
  int olen = strlen(s);
  char *temp = (char *) malloc(olen + 1);
  strcpy(temp, s);
  nstr = strtok(temp, sep);
  
loop:  

  if(nstr != NULL)
  {
    x.push(nstr);
    nstr = strtok(NULL, sep);
    goto loop;
  }

  return x;
}


//===== files


solarray file(text name) { return file(name.data); }
solarray file(const char *name)
{
 FILE *f;
 char BUFFER[65536];
 char *ptr;
 solarray a = solarray();

 f = fopen(name, "r");
 if(f == NULL) return solarray();
 while(1)
 {
  ptr = fgets(BUFFER, 65536, f);
  if(ptr == NULL) break;
  a.push(dyn(ptr));
 }
 fclose(f);
 return a;
}

void store(solarray a, text name) { store(a, name.data); }
void store(solarray a, const char *name)
{
 FILE *f;

 f = fopen(name, "w");
 if(f == NULL) return;
 dyn x;
 a.begin();
 do
 {
   x.set(a.current());
   fputs(x.toStr(), f);
 }
 while(!dyn::isNil(a.inc()));

 fclose(f);
}

char *extractProperty(const char *t, const char *prop)
{
  const char *x;
  int pos = text(t).findLex(prop);
  if(pos == -1) return "";
  pos += strlen(prop);

  x = t + pos;
  while(*x == ' ' || *x == '=' || *x == '\'' || *x == '\"') x++;
  pos = 0;
  while(x[pos] != '\'' && x[pos] != '\"' && x[pos] != '>') pos++;
  return substr(x, 0, pos);
}

dict get_meta_tags(const char *fname) {
  solarray a = file(fname);
  dict b = dict();
  int pos, eot;
  bool inside = false;
  text t, metas;

  a.begin();
  do
  {
   t = a.current().toText();
   if(!inside)
   {
     pos = t.findLex("<head>");
     if(pos != -1)
     {
       t = t.slice(pos + 6, t.length(), 0);
       inside = true;
     }
   }
   if(inside)
   {
     pos = t.findLex("</head>");
     if(pos != -1)
     {
       b.push(t.slice(0, pos, 0));
       break;
     }
     b.push(t.trim());
   }
  } while(!dyn::isNil(a.inc()));

  t = b.join("");

// now search metas inside the text

  a.clear();

  pos = 0;
  while(pos != -1)
  {
    pos = t.findLex("<meta", pos);
    if(pos != -1)
    {
      eot = t.find(">", pos + 5);
      if(eot != -1)
      {
        a.push(dyn(t.slice(pos, eot, 1)));
      }
      pos = eot;
    }
  }

 b.clear();
 for(int i = 0; i < a.size(); i++)
 {
   t = a[i].toText();
   metas = extractProperty(t.data, "name");
   if(metas.len > 0)
    b[metas] = extractProperty(t.data, "content");
 }

 return b;
}


solarray array_reverse(solarray x, bool preserve)
{
  solarray y;
  int len = x.size() - 1;
  int middle = len / 2;
  y.set(x);
  for(int i = 0; i < middle; i++)
  {
    if (preserve == true) y.swapElem(i, len - i);
                     else y.swapData(i, len - i);
  }
  return y;
}

// mut the content of a dir into an solarray

solarray _scandir(DIR *x)	
{
	solarray a = solarray();

	while(1)
	{	
		struct dirent *de = readdir(x);
		if(de == NULL) break;	
		char *str = de->d_name;	
		if(str != NULL)
		{	
			if(strlen(str) > 0)
			{	
				if( str[0] != '.')  
				{	
					a.push(dyn(str));
				}	
				continue;
			}	
		}	
		break;
	}
	return a;	
}	


char *str(solarray a)
{
	return a.toString();
}
